
/*
 * de.unkrig.commons - A general-purpose Java class library
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.commons.file.filetransformation;

import static de.unkrig.commons.lang.ExceptionUtil.wrap;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import de.unkrig.commons.file.ExceptionHandler;
import de.unkrig.commons.file.contentstransformation.ContentsTransformer;
import de.unkrig.commons.file.contentstransformation.GzipContentsTransformer;
import de.unkrig.commons.file.contentstransformation.ZipContentsTransformer;
import de.unkrig.commons.file.zipentrytransformation.RenameZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.SelectiveZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntriesTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryContentsTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformer;
import de.unkrig.commons.file.zipentrytransformation.ZipEntryTransformerUtil;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.pattern.Glob;

public
class FileTransformationUtil {

    private FileTransformationUtil() {}

    /**
     * @see #transform(List, File, FileTransformer, ExceptionHandler)
     */
    public static void
    transform(
        List<File>      inputFiles,
        @Nullable File  outputFile,
        FileTransformer fileTransformer
    ) throws IOException {
        transform(inputFiles, outputFile, fileTransformer, ExceptionHandler.<IOException>defaultHandler());
    }

    /**
     * If {@code outputFile} is {@code null}, then the {@code inputFiles} are transformed in-place.
     * <p>
     * Otherwise, if the {@code outputFile} is an existing directory, then the {@code inputFiles} are transformed into
     * files with the same names in the {@code outputFile} directory.
     * <p>
     * Otherwise, then number of {@code inputFiles} must be one, and the single input file is transformed into the
     * {@code outputFile}.
     *
     * @param exceptionHandler If the transformation of one of the {@code inputFiles} throws an {@link IOException} or
     *                         a {@link RuntimeException}, then {@link ExceptionHandler#handle(Exception)} resp. {@link
     *                         ExceptionHandler#handle(RuntimeException)} is called. Iff that call completes normally,
     *                         then processing continues with the next {@code inputFile}.
     */
    public static void
    transform(
        List<File>                    inputFiles,
        @Nullable File                outputFile,
        FileTransformer               fileTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) throws IOException {
        if (outputFile == null) {
            for (File inputFile : inputFiles) {
                try {
                    fileTransformer.transform(inputFile, inputFile);
                } catch (IOException ioe) {
                    exceptionHandler.handle(wrap("Transforming '" + inputFile + "'", ioe));
                } catch (RuntimeException re) {
                    exceptionHandler.handle(wrap("Transforming '" + inputFile + "'", re));
                }
            }
        } else
        if (outputFile.isDirectory()) {
            for (File inputFile : inputFiles) {
                try {
                    fileTransformer.transform(inputFile, new File(outputFile, inputFile.getName()));
                } catch (IOException ioe) {
                    exceptionHandler.handle(wrap("Transforming '" + inputFile + "' into '" + outputFile + "'", ioe));
                } catch (RuntimeException re) {
                    exceptionHandler.handle(wrap("Transforming '" + inputFile + "' into '" + outputFile + "'", re));
                }
            }
        } else
        {
            if (inputFiles.size() != 1) {
                throw new IOException("Exactly input file must be specified because '" + outputFile + "' is not an existing directory");
            }
            fileTransformer.transform(inputFiles.get(0), outputFile);
        }
    }

    /**
     * Creates a file transformer that transforms each file's contents (optionally looking into ZIP and GZIP format
     * files, recursively, renaming and/or removing entries on-the-fly) through the given contents transformer.
     * <pre>
     *                 |
     *                 v
     *       InPlaceFileTransformer
     *                 |
     *                 v
     * +-----+  ZipFileTransformer
     * |     |   |entry     |default
     * |     v   v          v
     * |   RemoveZET      FileCT
     * |       |            |
     * |       v            |
     * |   RenameZET        |
     * |       |            |
     * |       v            |
     * |   ZipEntryCT       |
     * |       |            |
     * |       v            v
     * |     GzipCT       GzipCT
     * |       |   contents| |default
     * |       | +---------+ |
     * |       v v           |
     * |      ZipCT          |
     * |  entry| |default    |
     * +-------+ +---------+ |
     *                     | |
     *                     v v
     *                  defaultCT
     * </pre>
     */
    public static FileTransformer
    newZipFileTransformer(
        boolean                       lookIntoZip,
        Predicate<String>             zipEntryRemoval,
        Glob                          zipEntryRenaming,
        ZipEntriesTransformer         zipEntriesTransformer,
        boolean                       keepOriginals,
        ContentsTransformer           defaultContentsTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        if (!lookIntoZip) return new FileContentsTransformer(defaultContentsTransformer);

        final ZipEntryTransformer[] zipEntryTransformer = new ZipEntryTransformer[1];

        ZipContentsTransformer zipContentsTransformer = new ZipContentsTransformer(
            zipEntriesTransformer,
            new ZipEntryTransformer() {

                @Override public void
                transform(ZipEntry entry, String path, InputStream is, ZipOutputStream zos) throws IOException {
                    zipEntryTransformer[0].transform(entry, path, is, zos);
                }
            },
            defaultContentsTransformer,
            exceptionHandler
        );

        FileTransformer fileTransformer = new FileContentsTransformer(new GzipContentsTransformer(zipContentsTransformer, defaultContentsTransformer));

        zipEntryTransformer[0] = new ZipEntryContentsTransformer(new GzipContentsTransformer(zipContentsTransformer, zipContentsTransformer));

        zipEntryTransformer[0] = new RenameZipEntryTransformer(zipEntryRenaming, zipEntryTransformer[0]);

        zipEntryTransformer[0] = new SelectiveZipEntryTransformer(zipEntryRemoval, ZipEntryTransformerUtil.nop(), zipEntryTransformer[0]);

        fileTransformer = new ZipFileTransformer(
            fileTransformer,       // defaultFileTransformer
            zipEntriesTransformer, // zipEntriesTransformer
            zipEntryTransformer[0] // zipEntryTransformer
        );

        fileTransformer = new InPlaceFileTransformer(fileTransformer, keepOriginals);

        return fileTransformer;
    }

    /**
     * Creates a file transformer that removes certain files, renames certain files, and then transforms each file's
     * contents (recursing through directories) through the given contents transformer.
     * <pre>
     *          |
     *          |   +---------+
     *          |   |         |
     *          v   v         |
     *        RemoveFT        |
     *            |           |
     *            v           |
     *        RenameFT        |
     *            |           |
     *            v           |
     *  DirectoryTransformer  |
     *   default|  |member    |
     *          |  +----------+
     *          v
     *   fileTransformer
     * </pre>
     */
    public static FileTransformer
    newDirectoryTreeTransformer(
        Predicate<String>             memberRemoval,
        Glob                          memberRenaming,
        final FileAdder               memberAdder,
        boolean                       keepOriginals,
        FileTransformer               fileTransformer,
        ExceptionHandler<IOException> exceptionHandler
    ) {
        final FileTransformer[] memberTransformer = new FileTransformer[1];

        memberTransformer[0] = new DirectoryTransformer(
            fileTransformer,                    // defaultFileTransformer
            new DirectoryMembersTransformer() { // directoryMembersTransformer

                @Override public void
                visitMember(File inputFile, File outputFile, FileTransformer fileTransformer) throws IOException {
                    fileTransformer.transform(inputFile, outputFile);
                }

                @Override public void
                visitEnd(File inputDirectory, File outputDirectory) throws IOException {
                    memberAdder.addFiles(outputDirectory);
                }
            },
            new FileTransformer() {             // memberTransformer

                @Override public void
                transform(File in, File out) throws IOException {
                    memberTransformer[0].transform(in, out);
                }
            },
            exceptionHandler                    // exceptionHandler
        );

        memberTransformer[0] = new RenameFileTransformer(memberRenaming, memberTransformer[0], keepOriginals);

        memberTransformer[0] = new SelectiveFileTransformer(memberRemoval, new RemoveFileTransformer(keepOriginals), memberTransformer[0]);

        return memberTransformer[0];
    }
}
