
/*
 * de.unkrig.loggifier - Injects logging code into JAVA[TM] class files
 *
 * Copyright (c) 2011, Arno Unkrig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
 * following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *       following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
 *       following disclaimer in the documentation and/or other materials provided with the distribution.
 *    3. The name of the author may not be used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.unkrig.loggifier;

import static de.unkrig.commons.text.pattern.Glob.INCLUDES_EXCLUDES;
import static de.unkrig.commons.text.pattern.Pattern2.WILDCARD;
import static java.util.logging.Level.SEVERE;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.ResourceCollection;
import org.apache.tools.ant.types.resources.FileResource;

import de.unkrig.commons.file.filetransformation.FileTransformationUtil;
import de.unkrig.commons.lang.protocol.Predicate;
import de.unkrig.commons.lang.protocol.PredicateUtil;
import de.unkrig.commons.nullanalysis.Nullable;
import de.unkrig.commons.text.pattern.Glob;
import de.unkrig.loggifier.LoggifyingClassAdapter.ErrorHandler;

/**
 * An ANT task for loggification of Java class files.
 *
 * @see Loggifier
 */
public
class AntTask extends Task {

    static final Logger LOGGER = Logger.getLogger(AntTask.class.getName());

    private final RuleSet     ruleSet = new RuleSet(LoggifyingContentsTransformer.DEFAULT_LOG_LEVEL_CALCULATOR);
    private final List<File>  inputFiles = new ArrayList<File>();
    private @Nullable File    outputFile = null;
    private Predicate<String> selector = PredicateUtil.always();
    private boolean           keepOriginals;

    // ANT task attribute setters.

    public void
    setSelector(String value) {
        this.selector = Glob.compile(value, WILDCARD | INCLUDES_EXCLUDES);
    }

    public void
    setOut(File value) {
        this.outputFile = value;
    }

    public void
    setKeepOriginals(boolean value) {
        this.keepOriginals = value;
    }

    // ANT subelement adders.

    public static
    class ElementWithText {
        String text = "";
        public void addText(String text) { this.text = text; }
    }

    public void
    addConfiguredRule(ElementWithText element) {
        try {
            this.ruleSet.add(element.text);
        } catch (Exception e) {
            LOGGER.log(SEVERE, e.toString(), e);
            throw new BuildException(e);
        }
    }

    public void
    addConfigured(ResourceCollection files) {
        if (!files.isFilesystemOnly()) throw new IllegalArgumentException("'" + files + "' is not on filesystem");

        for (@SuppressWarnings("unchecked") Iterator<FileResource> it = files.iterator(); it.hasNext();) {
            this.inputFiles.add(it.next().getFile());
        }
    }

    /**
     * ANT execute method.
     */
    @Override public void
    execute() throws BuildException {
        try {
            final int[] errorCount = new int[1];
            ErrorHandler errorHandler = new LoggifyingClassAdapter.ErrorHandler() {

                @Override public void
                handle(String message) {
                    System.err.println(message);
                    errorCount[0]++;
                }

                @Override public void
                handle(Throwable t) {
                    t.printStackTrace();
                    errorCount[0]++;
                }
            };
            Loggifier loggifier = new Loggifier(this.selector, this.ruleSet, this.keepOriginals, errorHandler);
            FileTransformationUtil.transform(
                this.inputFiles,
                this.outputFile,
                loggifier
            );

            if (errorCount[0] > 0) throw new BuildException(errorCount[0] + " error(s) while loggifying");
        } catch (Exception e) {
            LOGGER.log(SEVERE, e.toString(), e);
            throw new BuildException(e);
        }
    }
}
